/// <reference path="mwg.d.ts" />
declare module org {
    module mwg {
        module structure {
            class StructurePlugin implements org.mwg.plugin.Plugin {
                start(graph: org.mwg.Graph): void;
                stop(): void;
            }
            interface Tree extends org.mwg.Node {
                setDistance(distanceType: number): void;
                insert(keys: Float64Array, value: number): void;
                profile(keys: Float64Array, occurrence: number): void;
                nearestN(keys: Float64Array, nbElem: number): org.mwg.structure.TreeResult;
                nearestWithinRadius(keys: Float64Array, radius: number): org.mwg.structure.TreeResult;
                nearestNWithinRadius(keys: Float64Array, nbElem: number, radius: number): org.mwg.structure.TreeResult;
                query(min: Float64Array, max: Float64Array): org.mwg.structure.TreeResult;
                size(): number;
                numberOfNodes(): number;
            }
            interface TreeResult {
                size(): number;
                insert(key: Float64Array, value: number, distance: number): boolean;
                keys(index: number): Float64Array;
                value(index: number): number;
                distance(index: number): number;
                getWorstDistance(): number;
                free(): void;
                isCapacityReached(): boolean;
                sort(ascending: boolean): void;
            }
            module action {
                class NearestNWithinRadius implements org.mwg.task.Action {
                    static NAME: string;
                    private _key;
                    private _n;
                    private _radius;
                    private _fetchNodes;
                    constructor(n: number, radius: number, key: Float64Array, fetchNodes: boolean);
                    eval(ctx: org.mwg.task.TaskContext): void;
                    serialize(builder: java.lang.StringBuilder): void;
                    toString(): string;
                }
            }
            module distance {
                class CosineDistance implements org.mwg.structure.distance.Distance {
                    private static static_instance;
                    static instance(): org.mwg.structure.distance.CosineDistance;
                    constructor();
                    measure(x: Float64Array, y: Float64Array): number;
                    compare(x: number, y: number): boolean;
                    getMinValue(): number;
                    getMaxValue(): number;
                }
                interface Distance {
                    measure(x: Float64Array, y: Float64Array): number;
                    compare(x: number, y: number): boolean;
                    getMinValue(): number;
                    getMaxValue(): number;
                }
                class Distances {
                    static EUCLIDEAN: number;
                    static GEODISTANCE: number;
                    static COSINE: number;
                    static PEARSON: number;
                    static DEFAULT: number;
                    static getDistance(distance: number): org.mwg.structure.distance.Distance;
                }
                class EuclideanDistance implements org.mwg.structure.distance.Distance {
                    private static static_instance;
                    static instance(): org.mwg.structure.distance.EuclideanDistance;
                    constructor();
                    measure(x: Float64Array, y: Float64Array): number;
                    compare(x: number, y: number): boolean;
                    getMinValue(): number;
                    getMaxValue(): number;
                }
                class GeoDistance implements org.mwg.structure.distance.Distance {
                    private static static_instance;
                    static instance(): org.mwg.structure.distance.GeoDistance;
                    constructor();
                    measure(x: Float64Array, y: Float64Array): number;
                    private static toRadians(angledeg);
                    compare(x: number, y: number): boolean;
                    getMinValue(): number;
                    getMaxValue(): number;
                }
                class PearsonDistance implements org.mwg.structure.distance.Distance {
                    private static static_instance;
                    static instance(): org.mwg.structure.distance.PearsonDistance;
                    constructor();
                    measure(a: Float64Array, b: Float64Array): number;
                    compare(x: number, y: number): boolean;
                    getMinValue(): number;
                    getMaxValue(): number;
                }
            }
            module trees {
                class IndexStrategy {
                    static INDEX: number;
                    static PROFILE: number;
                    static DEFAULT: number;
                }
                class KDTree extends org.mwg.base.BaseNode implements org.mwg.structure.Tree {
                    static NAME: string;
                    static RESOLUTION: string;
                    static DISTANCE: string;
                    private static EGRAPH;
                    private static STRATEGY;
                    private static DIM;
                    static DISTANCE_DEF: number;
                    private static E_SUBTREE_NODES;
                    private static E_KEY;
                    private static E_SUM_KEY;
                    private static E_VALUE;
                    private static E_SUBTREE_VALUES;
                    private static E_RIGHT;
                    private static E_LEFT;
                    constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph);
                    private static checkCreateLevels(key1, key2, resolutions);
                    static rangeSearch(lowk: Float64Array, uppk: Float64Array, center: Float64Array, distance: org.mwg.structure.distance.Distance, node: org.mwg.struct.ENode, lev: number, dim: number, nnl: org.mwg.structure.util.VolatileResult): void;
                    private static reccursiveTraverse(node, nnl, distance, target, hr, lev, dim, max_dist_sqd, radius);
                    private static internalInsert(node, key, value, strategyType, lev, resolution);
                    private static createNode(parent, right);
                    setDistance(distanceType: number): void;
                    insert(keys: Float64Array, value: number): void;
                    profile(keys: Float64Array, occurrence: number): void;
                    nearestN(keys: Float64Array, nbElem: number): org.mwg.structure.TreeResult;
                    nearestWithinRadius(keys: Float64Array, radius: number): org.mwg.structure.TreeResult;
                    nearestNWithinRadius(keys: Float64Array, nbElem: number, radius: number): org.mwg.structure.TreeResult;
                    query(min: Float64Array, max: Float64Array): org.mwg.structure.TreeResult;
                    size(): number;
                    numberOfNodes(): number;
                }
                class NDTree extends org.mwg.base.BaseNode implements org.mwg.structure.Tree {
                    static NAME: string;
                    static BOUND_MIN: string;
                    static BOUND_MAX: string;
                    static RESOLUTION: string;
                    static BUFFER_SIZE: string;
                    static DISTANCE: string;
                    private static EGRAPH;
                    private static STRATEGY;
                    static BUFFER_SIZE_DEF: number;
                    static DISTANCE_DEF: number;
                    private static E_TOTAL;
                    private static E_SUBNODES;
                    private static E_TOTAL_SUBNODES;
                    private static E_MIN;
                    private static E_MAX;
                    private static E_BUFFER_KEYS;
                    private static E_BUFFER_VALUES;
                    private static E_VALUE;
                    private static E_PROFILE;
                    private static E_OFFSET_REL;
                    constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph);
                    private static getRelationId(centerKey, keyToInsert);
                    private static checkCreateLevels(min, max, resolutions);
                    private static getCenterMinMax(min, max);
                    private static getCenter(node);
                    private static check(values, min, max);
                    private static createNewNode(parent, root, index, min, max, center, keyToInsert, buffersize);
                    private static subInsert(parent, key, value, strategyType, min, max, center, resolution, buffersize, root, bufferupdate);
                    private static internalInsert(node, key, value, strategyType, min, max, center, resolution, buffersize, root);
                    private static compare(key1, key2, resolution);
                    setDistance(distanceType: number): void;
                    insert(keys: Float64Array, value: number): void;
                    profile(keys: Float64Array, occurrence: number): void;
                    nearestN(keys: Float64Array, nbElem: number): org.mwg.structure.TreeResult;
                    nearestWithinRadius(keys: Float64Array, radius: number): org.mwg.structure.TreeResult;
                    nearestNWithinRadius(keys: Float64Array, nbElem: number, radius: number): org.mwg.structure.TreeResult;
                    query(min: Float64Array, max: Float64Array): org.mwg.structure.TreeResult;
                    size(): number;
                    numberOfNodes(): number;
                    private static binaryFromLong(value, dim);
                    private static reccursiveTraverse(node, calcZone, nnl, strategyType, distance, target, targetmin, targetmax, targetcenter, radius);
                }
            }
            module util {
                class HRect {
                    min: Float64Array;
                    max: Float64Array;
                    constructor(vmin: Float64Array, vmax: Float64Array);
                    clone(): any;
                    closest(t: Float64Array): Float64Array;
                    static infiniteHRect(d: number): org.mwg.structure.util.HRect;
                    toString(): string;
                }
                class TreeHelper {
                    static checkBoundsIntersection(targetmin: Float64Array, targetmax: Float64Array, boundMin: Float64Array, boundMax: Float64Array): boolean;
                    static checkKeyInsideBounds(key: Float64Array, boundMin: Float64Array, boundMax: Float64Array): boolean;
                    static getclosestDistance(target: Float64Array, boundMin: Float64Array, boundMax: Float64Array, distance: org.mwg.structure.distance.Distance): number;
                    static filterAndInsert(key: Float64Array, value: number, target: Float64Array, targetmin: Float64Array, targetmax: Float64Array, targetcenter: Float64Array, distance: org.mwg.structure.distance.Distance, radius: number, nnl: org.mwg.structure.util.VolatileResult): void;
                }
                class VolatileResult implements org.mwg.structure.TreeResult {
                    private static maxPriority;
                    private static _KEYS;
                    private static _VALUES;
                    private static _DISTANCES;
                    private node;
                    private capacity;
                    private count;
                    private worst;
                    private _keys;
                    private _values;
                    private _distances;
                    constructor(node: org.mwg.struct.ENode, capacity: number);
                    size(): number;
                    insert(key: Float64Array, value: number, distance: number): boolean;
                    private add(key, value, distance, remove);
                    private remove();
                    private bubbleUp(pos);
                    private bubbleDown(pos);
                    keys(index: number): Float64Array;
                    value(index: number): number;
                    distance(index: number): number;
                    getWorstDistance(): number;
                    free(): void;
                    isCapacityReached(): boolean;
                    private swap(i, j);
                    private partition(l, h, ascending);
                    private quickSort(l, h, ascending);
                    sort(ascending: boolean): void;
                }
            }
        }
    }
}
