/// <reference path="mwg.d.ts" />

module org {
  export module mwg {
    export module structure {
      export class StructurePlugin implements org.mwg.plugin.Plugin {
        public start(graph: org.mwg.Graph): void {
          graph.actionRegistry().declaration(org.mwg.structure.action.NearestNWithinRadius.NAME).setParams(org.mwg.Type.INT, org.mwg.Type.DOUBLE, org.mwg.Type.DOUBLE_ARRAY).setFactory((params : any[]) => {
{
              return new org.mwg.structure.action.NearestNWithinRadius(<number>params[0], <number>params[1], <Float64Array>params[2], <boolean>params[3]);
            }          });
          graph.nodeRegistry().declaration(org.mwg.structure.trees.KDTree.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.structure.trees.KDTree(world, time, id, graph);
            }          });
          graph.nodeRegistry().declaration(org.mwg.structure.trees.NDTree.NAME).setFactory((world : number, time : number, id : number, graph : org.mwg.Graph) => {
{
              return new org.mwg.structure.trees.NDTree(world, time, id, graph);
            }          });
        }
        public stop(): void {}
      }
      export interface Tree extends org.mwg.Node {
        setDistance(distanceType: number): void;
        insert(keys: Float64Array, value: number): void;
        profile(keys: Float64Array, occurrence: number): void;
        nearestN(keys: Float64Array, nbElem: number): org.mwg.structure.TreeResult;
        nearestWithinRadius(keys: Float64Array, radius: number): org.mwg.structure.TreeResult;
        nearestNWithinRadius(keys: Float64Array, nbElem: number, radius: number): org.mwg.structure.TreeResult;
        query(min: Float64Array, max: Float64Array): org.mwg.structure.TreeResult;
        size(): number;
        numberOfNodes(): number;
      }
      export interface TreeResult {
        size(): number;
        insert(key: Float64Array, value: number, distance: number): boolean;
        keys(index: number): Float64Array;
        value(index: number): number;
        distance(index: number): number;
        getWorstDistance(): number;
        free(): void;
        isCapacityReached(): boolean;
        sort(ascending: boolean): void;
      }
      export module action {
        export class NearestNWithinRadius implements org.mwg.task.Action {
          public static NAME: string = "NearestNWithinRadius";
          private _key: Float64Array;
          private _n: number;
          private _radius: number;
          private _fetchNodes: boolean;
          constructor(n: number, radius: number, key: Float64Array, fetchNodes: boolean) {
            this._key = key;
            this._n = n;
            this._radius = radius;
            this._fetchNodes = fetchNodes;
          }
          public eval(ctx: org.mwg.task.TaskContext): void {
            let previousResult: org.mwg.task.TaskResult<any> = ctx.result();
            let nextResult: org.mwg.task.TaskResult<any> = ctx.newResult();
            if (previousResult != null) {
              let defer: org.mwg.DeferCounter = ctx.graph().newCounter(previousResult.size());
              let previousResultIt: org.mwg.task.TaskResultIterator<any> = previousResult.iterator();
              let iter: any = previousResultIt.next();
              while (iter != null) {
                if (iter instanceof org.mwg.structure.trees.NDTree || iter instanceof org.mwg.structure.trees.KDTree) {
                  let tr: org.mwg.structure.TreeResult = (<org.mwg.structure.Tree>iter).nearestNWithinRadius(this._key, this._n, this._radius);
                  if (this._fetchNodes) {
                    let nodeIds: Float64Array = new Float64Array(tr.size());
                    for (let i: number = 0; i < tr.size(); i++) {
                      nodeIds[i] = tr.value(i);
                    }
                    ctx.graph().lookupAll(ctx.world(), ctx.time(), nodeIds, (result)=>{
                      for (let j: number = 0; j < result.length; j++) {
                        let line: org.mwg.task.TaskResult<any> = ctx.newResult();
                        line.add(tr.keys(j));
                        line.add(result[j]);
                        line.add(tr.distance(j));
                        nextResult.add(line);
                      }
                      tr.free();
                      defer.count();
                    });
                  } else {
                    for (let i: number = 0; i < tr.size(); i++) {
                      let line: org.mwg.task.TaskResult<any> = ctx.newResult();
                      line.add(tr.keys(i));
                      line.add(tr.value(i));
                      line.add(tr.distance(i));
                      nextResult.add(line);
                    }
                    tr.free();
                    defer.count();
                  }
                } else {
                  defer.count();
                }
                iter = previousResultIt.next();
              }
              defer.then(() => {
{
                  ctx.continueWith(nextResult);
                }              });
            } else {
              ctx.continueWith(nextResult);
            }
          }
          public serialize(builder: java.lang.StringBuilder): void {
            throw new Error("Not supported yet!");
          }
          public toString(): string {
            return "NearestNWithinRadius(\'" + "\')";
          }
        }
      }
      export module distance {
        export class CosineDistance implements org.mwg.structure.distance.Distance {
          private static static_instance: org.mwg.structure.distance.CosineDistance = null;
          public static instance(): org.mwg.structure.distance.CosineDistance {
            if (CosineDistance.static_instance == null) {
              CosineDistance.static_instance = new org.mwg.structure.distance.CosineDistance();
            }
            return CosineDistance.static_instance;
          }
          constructor() {}
          public measure(x: Float64Array, y: Float64Array): number {
            let sumTop: number = 0;
            let sumOne: number = 0;
            let sumTwo: number = 0;
            for (let i: number = 0; i < x.length; i++) {
              sumTop += x[i] * y[i];
              sumOne += x[i] * x[i];
              sumTwo += y[i] * y[i];
            }
            let cosSim: number = sumTop / (Math.sqrt(sumOne) * Math.sqrt(sumTwo));
            if (cosSim < 0) {
              cosSim = 0;
            }
            return 1 - cosSim;
          }
          public compare(x: number, y: number): boolean {
            return x < y;
          }
          public getMinValue(): number {
            return 0;
          }
          public getMaxValue(): number {
            return java.lang.Double.MAX_VALUE;
          }
        }
        export interface Distance {
          measure(x: Float64Array, y: Float64Array): number;
          compare(x: number, y: number): boolean;
          getMinValue(): number;
          getMaxValue(): number;
        }
        export class Distances {
          public static EUCLIDEAN: number = 0;
          public static GEODISTANCE: number = 1;
          public static COSINE: number = 2;
          public static PEARSON: number = 3;
          public static DEFAULT: number = Distances.EUCLIDEAN;
          public static getDistance(distance: number): org.mwg.structure.distance.Distance {
            switch (distance) {
              case Distances.EUCLIDEAN:
                return org.mwg.structure.distance.EuclideanDistance.instance();
              case Distances.GEODISTANCE:
                return org.mwg.structure.distance.GeoDistance.instance();
              case Distances.COSINE:
                return org.mwg.structure.distance.CosineDistance.instance();
              case Distances.PEARSON:
                return org.mwg.structure.distance.PearsonDistance.instance();
            }
            return org.mwg.structure.distance.Distances.getDistance(Distances.DEFAULT);
          }
        }
        export class EuclideanDistance implements org.mwg.structure.distance.Distance {
          private static static_instance: org.mwg.structure.distance.EuclideanDistance = null;
          public static instance(): org.mwg.structure.distance.EuclideanDistance {
            if (EuclideanDistance.static_instance == null) {
              EuclideanDistance.static_instance = new org.mwg.structure.distance.EuclideanDistance();
            }
            return EuclideanDistance.static_instance;
          }
          constructor() {}
          public measure(x: Float64Array, y: Float64Array): number {
            let value: number = 0;
            for (let i: number = 0; i < x.length; i++) {
              value = value + (x[i] - y[i]) * (x[i] - y[i]);
            }
            return Math.sqrt(value);
          }
          public compare(x: number, y: number): boolean {
            return x < y;
          }
          public getMinValue(): number {
            return 0;
          }
          public getMaxValue(): number {
            return java.lang.Double.MAX_VALUE;
          }
        }
        export class GeoDistance implements org.mwg.structure.distance.Distance {
          private static static_instance: org.mwg.structure.distance.GeoDistance = null;
          public static instance(): org.mwg.structure.distance.GeoDistance {
            if (GeoDistance.static_instance == null) {
              GeoDistance.static_instance = new org.mwg.structure.distance.GeoDistance();
            }
            return GeoDistance.static_instance;
          }
          constructor() {}
          public measure(x: Float64Array, y: Float64Array): number {
            let earthRadius: number = 6371000;
            let dLat: number = org.mwg.structure.distance.GeoDistance.toRadians(y[0] - x[0]);
            let dLng: number = org.mwg.structure.distance.GeoDistance.toRadians(y[1] - x[1]);
            let a: number = Math.sin(dLat / 2) * Math.sin(dLat / 2) + Math.cos(org.mwg.structure.distance.GeoDistance.toRadians(x[0])) * Math.cos(org.mwg.structure.distance.GeoDistance.toRadians(y[0])) * Math.sin(dLng / 2) * Math.sin(dLng / 2);
            let c: number = 2 * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
            return earthRadius * c;
          }
          private static toRadians(angledeg: number): number {
            return angledeg * Math.PI / 180;
          }
          public compare(x: number, y: number): boolean {
            return x < y;
          }
          public getMinValue(): number {
            return 0;
          }
          public getMaxValue(): number {
            return java.lang.Double.MAX_VALUE;
          }
        }
        export class PearsonDistance implements org.mwg.structure.distance.Distance {
          private static static_instance: org.mwg.structure.distance.PearsonDistance = null;
          public static instance(): org.mwg.structure.distance.PearsonDistance {
            if (PearsonDistance.static_instance == null) {
              PearsonDistance.static_instance = new org.mwg.structure.distance.PearsonDistance();
            }
            return PearsonDistance.static_instance;
          }
          constructor() {}
          public measure(a: Float64Array, b: Float64Array): number {
            let xy: number = 0, x: number = 0, x2: number = 0, y: number = 0, y2: number = 0;
            for (let i: number = 0; i < a.length; i++) {
              xy += a[i] * b[i];
              x += a[i];
              y += b[i];
              x2 += a[i] * a[i];
              y2 += b[i] * b[i];
            }
            let n: number = a.length;
            return (xy - (x * y) / n) / Math.sqrt((x2 - (x * x) / n) * (y2 - (y * y) / n));
          }
          public compare(x: number, y: number): boolean {
            return Math.abs(x) > Math.abs(y);
          }
          public getMinValue(): number {
            return 1;
          }
          public getMaxValue(): number {
            return 0;
          }
        }
      }
      export module trees {
        export class IndexStrategy {
          public static INDEX: number = 0;
          public static PROFILE: number = 1;
          public static DEFAULT: number = IndexStrategy.INDEX;
        }
        export class KDTree extends org.mwg.base.BaseNode implements org.mwg.structure.Tree {
          public static NAME: string = "KDTree";
          public static RESOLUTION: string = "resolution";
          public static DISTANCE: string = "distance";
          private static EGRAPH: string = "egraph";
          private static STRATEGY: string = "strategy";
          private static DIM: string = "dim";
          public static DISTANCE_DEF: number = org.mwg.structure.distance.Distances.DEFAULT;
          private static E_SUBTREE_NODES: number = 0;
          private static E_KEY: number = 1;
          private static E_SUM_KEY: number = 2;
          private static E_VALUE: number = 3;
          private static E_SUBTREE_VALUES: number = 4;
          private static E_RIGHT: number = 5;
          private static E_LEFT: number = 6;
          constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
            super(p_world, p_time, p_id, p_graph);
          }
          private static checkCreateLevels(key1: Float64Array, key2: Float64Array, resolutions: Float64Array): boolean {
            if (resolutions != null) {
              for (let i: number = 0; i < key2.length; i++) {
                if (Math.abs(key1[i] - key2[i]) > resolutions[i]) {
                  return true;
                }
              }
            } else {
              for (let i: number = 0; i < key2.length; i++) {
                if ((key1[i] != key2[i])) {
                  return true;
                }
              }
            }
            return false;
          }
          public static rangeSearch(lowk: Float64Array, uppk: Float64Array, center: Float64Array, distance: org.mwg.structure.distance.Distance, node: org.mwg.struct.ENode, lev: number, dim: number, nnl: org.mwg.structure.util.VolatileResult): void {
            if (node == null)
              return;

            let key: Float64Array = <Float64Array>node.getAt(KDTree.E_KEY);
            if (lowk[lev] <= key[lev]) {
              org.mwg.structure.trees.KDTree.rangeSearch(lowk, uppk, center, distance, <org.mwg.struct.ENode>node.getAt(KDTree.E_LEFT), (lev + 1) % dim, dim, nnl);
            }
            let j: number;
            j = 0;
            while (j < dim && lowk[j] <= key[j] && uppk[j] >= key[j]) {
              j++;
            }
            if (j == dim) {
              nnl.insert(key, <number>node.getAt(KDTree.E_VALUE), distance.measure(key, center));
            }
            if (uppk[lev] > key[lev]) {
              org.mwg.structure.trees.KDTree.rangeSearch(lowk, uppk, center, distance, <org.mwg.struct.ENode>node.getAt(KDTree.E_RIGHT), (lev + 1) % dim, dim, nnl);
            }
          }
          private static reccursiveTraverse(node: org.mwg.struct.ENode, nnl: org.mwg.structure.util.VolatileResult, distance: org.mwg.structure.distance.Distance, target: Float64Array, hr: org.mwg.structure.util.HRect, lev: number, dim: number, max_dist_sqd: number, radius: number): void {
            if (node == null) {
              return;
            }
            let pivot: Float64Array = <Float64Array>node.getAt(KDTree.E_KEY);
            if (pivot == null) {
              throw new Error("Key can't be null");
            }
            let s: number = lev % dim;
            let pivot_to_target: number = distance.measure(pivot, target);
            let left_hr: org.mwg.structure.util.HRect = hr;
            let right_hr: org.mwg.structure.util.HRect = <org.mwg.structure.util.HRect>hr.clone();
            left_hr.max[s] = pivot[s];
            right_hr.min[s] = pivot[s];
            let target_in_left: boolean = target[s] < pivot[s];
            let nearer_kd: org.mwg.struct.ENode;
            let nearer_hr: org.mwg.structure.util.HRect;
            let further_kd: org.mwg.struct.ENode;
            let further_hr: org.mwg.structure.util.HRect;
            if (target_in_left) {
              nearer_kd = <org.mwg.struct.ENode>node.getAt(KDTree.E_LEFT);
              nearer_hr = left_hr;
              further_kd = <org.mwg.struct.ENode>node.getAt(KDTree.E_RIGHT);
              further_hr = right_hr;
            } else {
              nearer_kd = <org.mwg.struct.ENode>node.getAt(KDTree.E_RIGHT);
              nearer_hr = right_hr;
              further_kd = <org.mwg.struct.ENode>node.getAt(KDTree.E_LEFT);
              further_hr = left_hr;
            }
            org.mwg.structure.trees.KDTree.reccursiveTraverse(nearer_kd, nnl, distance, target, nearer_hr, lev + 1, dim, max_dist_sqd, radius);
            let dist_sqd: number;
            if (!nnl.isCapacityReached()) {
              dist_sqd = java.lang.Double.MAX_VALUE;
            } else {
              dist_sqd = nnl.getWorstDistance();
            }
            max_dist_sqd = Math.min(max_dist_sqd, dist_sqd);
            let closest: Float64Array = further_hr.closest(target);
            if (distance.measure(closest, target) < max_dist_sqd) {
              if (pivot_to_target < dist_sqd) {
                dist_sqd = pivot_to_target;
                if (radius > 0) {
                  if (dist_sqd < radius) {
                    nnl.insert(pivot, <number>node.getAt(KDTree.E_VALUE), dist_sqd);
                  }
                } else {
                  nnl.insert(pivot, <number>node.getAt(KDTree.E_VALUE), dist_sqd);
                }
                if (nnl.isCapacityReached()) {
                  max_dist_sqd = nnl.getWorstDistance();
                } else {
                  max_dist_sqd = java.lang.Double.MAX_VALUE;
                }
              }
              org.mwg.structure.trees.KDTree.reccursiveTraverse(further_kd, nnl, distance, target, further_hr, lev + 1, dim, max_dist_sqd, radius);
            }
          }
          private static internalInsert(node: org.mwg.struct.ENode, key: Float64Array, value: number, strategyType: number, lev: number, resolution: Float64Array): boolean {
            let pKey: Float64Array = <Float64Array>node.getAt(KDTree.E_KEY);
            if (pKey == null) {
              pKey = new Float64Array(key.length);
              java.lang.System.arraycopy(key, 0, pKey, 0, key.length);
              node.setAt(KDTree.E_KEY, org.mwg.Type.DOUBLE_ARRAY, pKey);
              node.setAt(KDTree.E_VALUE, org.mwg.Type.LONG, value);
              if (strategyType == org.mwg.structure.trees.IndexStrategy.PROFILE) {
                node.setAt(KDTree.E_SUBTREE_VALUES, org.mwg.Type.LONG, value);
                let sk: Float64Array = new Float64Array(pKey.length);
                for (let i: number = 0; i < key.length; i++) {
                  sk[i] = pKey[i] * value;
                }
                node.setAt(KDTree.E_SUM_KEY, org.mwg.Type.DOUBLE_ARRAY, sk);
              }
              node.setAt(KDTree.E_SUBTREE_NODES, org.mwg.Type.LONG, 1);
              return true;
            } else if (!org.mwg.structure.trees.KDTree.checkCreateLevels(key, pKey, resolution)) {
              if (strategyType == org.mwg.structure.trees.IndexStrategy.INDEX) {
                node.setAt(KDTree.E_VALUE, org.mwg.Type.LONG, value);
              } else if (strategyType == org.mwg.structure.trees.IndexStrategy.PROFILE) {
                let sk: Float64Array = <Float64Array>node.getAt(KDTree.E_SUM_KEY);
                for (let i: number = 0; i < pKey.length; i++) {
                  sk[i] = (sk[i] + key[i] * value);
                }
                node.setAt(KDTree.E_SUM_KEY, org.mwg.Type.DOUBLE_ARRAY, sk);
                node.setAt(KDTree.E_VALUE, org.mwg.Type.LONG, <number>node.getOrCreateAt(KDTree.E_VALUE, org.mwg.Type.LONG) + value);
                node.setAt(KDTree.E_SUBTREE_VALUES, org.mwg.Type.LONG, <number>node.getOrCreateAt(KDTree.E_SUBTREE_VALUES, org.mwg.Type.LONG) + value);
              }

              return false;
            } else {
              let child: org.mwg.struct.ENode;
              if (key[lev] > pKey[lev]) {
                child = <org.mwg.struct.ENode>node.getAt(KDTree.E_RIGHT);
                if (child == null) {
                  child = org.mwg.structure.trees.KDTree.createNode(node, true);
                }
              } else {
                child = <org.mwg.struct.ENode>node.getAt(KDTree.E_LEFT);
                if (child == null) {
                  child = org.mwg.structure.trees.KDTree.createNode(node, false);
                }
              }
              if (org.mwg.structure.trees.KDTree.internalInsert(child, key, value, strategyType, (lev + 1) % key.length, resolution)) {
                if (strategyType == org.mwg.structure.trees.IndexStrategy.PROFILE) {
                  node.setAt(KDTree.E_SUBTREE_VALUES, org.mwg.Type.LONG, <number>node.getAt(KDTree.E_SUBTREE_VALUES) + value);
                }
                node.setAt(KDTree.E_SUBTREE_NODES, org.mwg.Type.LONG, <number>node.getAt(KDTree.E_SUBTREE_NODES) + 1);
                return true;
              } else {
                if (strategyType == org.mwg.structure.trees.IndexStrategy.PROFILE) {
                  node.setAt(KDTree.E_SUBTREE_VALUES, org.mwg.Type.LONG, <number>node.getAt(KDTree.E_SUBTREE_VALUES) + value);
                }
                return false;
              }
            }

          }
          private static createNode(parent: org.mwg.struct.ENode, right: boolean): org.mwg.struct.ENode {
            let child: org.mwg.struct.ENode = parent.graph().newNode();
            if (right) {
              parent.setAt(KDTree.E_RIGHT, org.mwg.Type.ENODE, child);
            } else {
              parent.setAt(KDTree.E_LEFT, org.mwg.Type.ENODE, child);
            }
            return child;
          }
          public setDistance(distanceType: number): void {
            super.set(KDTree.DISTANCE, org.mwg.Type.INT, distanceType);
          }
          public insert(keys: Float64Array, value: number): void {
            let state: org.mwg.plugin.NodeState = this.unphasedState();
            let strategy: number = org.mwg.structure.trees.IndexStrategy.INDEX;
            let resolution: Float64Array = <Float64Array>state.getFromKey(KDTree.RESOLUTION);
            let graph: org.mwg.struct.EGraph = <org.mwg.struct.EGraph>state.getOrCreateFromKey(KDTree.EGRAPH, org.mwg.Type.EGRAPH);
{
              let root: org.mwg.struct.ENode = graph.root();
              if (root == null) {
                root = graph.newNode();
                state.setFromKey(KDTree.STRATEGY, org.mwg.Type.INT, strategy);
                graph.setRoot(root);
                state.setFromKey(KDTree.DIM, org.mwg.Type.INT, keys.length);
              } else {
                if (keys.length != <number>state.getFromKey(KDTree.DIM)) {
                  throw new Error("Keys should always be the same length");
                }
              }
              org.mwg.structure.trees.KDTree.internalInsert(root, keys, value, strategy, 0, resolution);
            }          }
          public profile(keys: Float64Array, occurrence: number): void {
            let state: org.mwg.plugin.NodeState = this.unphasedState();
            let strategy: number = org.mwg.structure.trees.IndexStrategy.PROFILE;
            let resolution: Float64Array = <Float64Array>state.getFromKey(KDTree.RESOLUTION);
            let graph: org.mwg.struct.EGraph = <org.mwg.struct.EGraph>state.getOrCreateFromKey(KDTree.EGRAPH, org.mwg.Type.EGRAPH);
{
              let root: org.mwg.struct.ENode = graph.root();
              if (root == null) {
                root = graph.newNode();
                state.setFromKey(KDTree.STRATEGY, org.mwg.Type.INT, strategy);
                graph.setRoot(root);
                state.setFromKey(KDTree.DIM, org.mwg.Type.INT, keys.length);
              } else {
                if (keys.length != <number>state.getFromKey(KDTree.DIM)) {
                  throw new Error("Keys should always be the same length");
                }
              }
              org.mwg.structure.trees.KDTree.internalInsert(root, keys, occurrence, strategy, 0, resolution);
            }          }
          public nearestN(keys: Float64Array, nbElem: number): org.mwg.structure.TreeResult {
            return this.nearestNWithinRadius(keys, -1, -1);
          }
          public nearestWithinRadius(keys: Float64Array, radius: number): org.mwg.structure.TreeResult {
            return this.nearestNWithinRadius(keys, -1, radius);
          }
          public nearestNWithinRadius(keys: Float64Array, nbElem: number, radius: number): org.mwg.structure.TreeResult {
            let state: org.mwg.plugin.NodeState = this.unphasedState();
            let graph: org.mwg.struct.EGraph = <org.mwg.struct.EGraph>state.getOrCreateFromKey(KDTree.EGRAPH, org.mwg.Type.EGRAPH);
            let distance: org.mwg.structure.distance.Distance = org.mwg.structure.distance.Distances.getDistance(state.getFromKeyWithDefault(KDTree.DISTANCE, KDTree.DISTANCE_DEF));
{
              let root: org.mwg.struct.ENode = graph.root();
              if (root == null) {
                return null;
              }
              if (keys.length != (<Float64Array>root.getAt(KDTree.E_KEY)).length) {
                throw new Error("Keys are not of the same size");
              }
              let calcZone: org.mwg.struct.EGraph = this.graph().space().newVolatileGraph();
              let nnl: org.mwg.structure.util.VolatileResult = new org.mwg.structure.util.VolatileResult(calcZone.newNode(), nbElem);
              org.mwg.structure.trees.KDTree.reccursiveTraverse(root, nnl, distance, keys, org.mwg.structure.util.HRect.infiniteHRect(keys.length), 0, keys.length, java.lang.Double.MAX_VALUE, radius);
              nnl.sort(true);
              return nnl;
            }          }
          public query(min: Float64Array, max: Float64Array): org.mwg.structure.TreeResult {
            let state: org.mwg.plugin.NodeState = this.unphasedState();
            let graph: org.mwg.struct.EGraph = <org.mwg.struct.EGraph>state.getOrCreateFromKey(KDTree.EGRAPH, org.mwg.Type.EGRAPH);
            let distance: org.mwg.structure.distance.Distance = org.mwg.structure.distance.Distances.getDistance(state.getFromKeyWithDefault(KDTree.DISTANCE, KDTree.DISTANCE_DEF));
            let center: Float64Array = new Float64Array(max.length);
            for (let i: number = 0; i < center.length; i++) {
              center[i] = (min[i] + max[i]) / 2;
            }
            let root: org.mwg.struct.ENode = graph.root();
            if (root == null) {
              return null;
            }
            let calcZone: org.mwg.struct.EGraph = this.graph().space().newVolatileGraph();
            let nnl: org.mwg.structure.util.VolatileResult = new org.mwg.structure.util.VolatileResult(calcZone.newNode(), -1);
            org.mwg.structure.trees.KDTree.rangeSearch(min, max, center, distance, root, 0, min.length, nnl);
            nnl.sort(true);
            return nnl;
          }
          public size(): number {
            return 0;
          }
          public numberOfNodes(): number {
            return 0;
          }
        }
        export class NDTree extends org.mwg.base.BaseNode implements org.mwg.structure.Tree {
          public static NAME: string = "NDTree";
          public static BOUND_MIN: string = "bound_min";
          public static BOUND_MAX: string = "bound_max";
          public static RESOLUTION: string = "resolution";
          public static BUFFER_SIZE: string = "buffer_size";
          public static DISTANCE: string = "distance";
          private static EGRAPH: string = "egraph";
          private static STRATEGY: string = "strategy";
          public static BUFFER_SIZE_DEF: number = 20;
          public static DISTANCE_DEF: number = org.mwg.structure.distance.Distances.DEFAULT;
          private static E_TOTAL: number = 0;
          private static E_SUBNODES: number = 1;
          private static E_TOTAL_SUBNODES: number = 2;
          private static E_MIN: number = 3;
          private static E_MAX: number = 4;
          private static E_BUFFER_KEYS: number = 5;
          private static E_BUFFER_VALUES: number = 6;
          private static E_VALUE: number = 7;
          private static E_PROFILE: number = 8;
          private static E_OFFSET_REL: number = 16;
          constructor(p_world: number, p_time: number, p_id: number, p_graph: org.mwg.Graph) {
            super(p_world, p_time, p_id, p_graph);
          }
          private static getRelationId(centerKey: Float64Array, keyToInsert: Float64Array): number {
            let result: number = 0;
            for (let i: number = 0; i < centerKey.length; i++) {
              if (i != 0) {
                result = result << 1;
              }
              if (keyToInsert[i] > centerKey[i]) {
                result += 1;
              }
            }
            return result + NDTree.E_OFFSET_REL;
          }
          private static checkCreateLevels(min: Float64Array, max: Float64Array, resolutions: Float64Array): boolean {
            if (resolutions != null) {
              for (let i: number = 0; i < min.length; i++) {
                if ((max[i] - min[i]) > 2 * resolutions[i]) {
                  return true;
                }
              }
            } else {
              for (let i: number = 0; i < min.length; i++) {
                if ((max[i] > min[i])) {
                  return true;
                }
              }
            }
            return false;
          }
          private static getCenterMinMax(min: Float64Array, max: Float64Array): Float64Array {
            let center: Float64Array = new Float64Array(min.length);
            for (let i: number = 0; i < min.length; i++) {
              center[i] = (max[i] + min[i]) / 2;
            }
            return center;
          }
          private static getCenter(node: org.mwg.struct.ENode): Float64Array {
            let min: Float64Array = <Float64Array>node.getAt(NDTree.E_MIN);
            let max: Float64Array = <Float64Array>node.getAt(NDTree.E_MAX);
            return org.mwg.structure.trees.NDTree.getCenterMinMax(min, max);
          }
          private static check(values: Float64Array, min: Float64Array, max: Float64Array): void {
            if (min == null || max == null) {
              throw new Error("Please set min and max boundary before inserting in the trees");
            }
            if (values.length != min.length) {
              throw new Error("Values dimension mismatch");
            }
            for (let i: number = 0; i < min.length; i++) {
              if (values[i] < min[i] || values[i] > max[i]) {
                throw new Error("Values should be between min, max!");
              }
            }
          }
          private static createNewNode(parent: org.mwg.struct.ENode, root: org.mwg.struct.ENode, index: number, min: Float64Array, max: Float64Array, center: Float64Array, keyToInsert: Float64Array, buffersize: number): org.mwg.struct.ENode {
            let node: org.mwg.struct.ENode = parent.graph().newNode();
            let minChild: Float64Array = new Float64Array(min.length);
            let maxChild: Float64Array = new Float64Array(max.length);
            for (let i: number = 0; i < min.length; i++) {
              if (keyToInsert[i] <= center[i]) {
                minChild[i] = min[i];
                maxChild[i] = center[i];
              } else {
                minChild[i] = center[i];
                maxChild[i] = max[i];
              }
            }
            node.setAt(NDTree.E_SUBNODES, org.mwg.Type.LONG, 0);
            node.setAt(NDTree.E_MIN, org.mwg.Type.DOUBLE_ARRAY, minChild);
            node.setAt(NDTree.E_MAX, org.mwg.Type.DOUBLE_ARRAY, maxChild);
            node.setAt(NDTree.E_TOTAL, org.mwg.Type.LONG, 0);
            root.setAt(NDTree.E_TOTAL_SUBNODES, org.mwg.Type.LONG, <number>root.getAt(NDTree.E_TOTAL_SUBNODES) + 1);
            parent.setAt(NDTree.E_SUBNODES, org.mwg.Type.LONG, <number>parent.getAt(NDTree.E_SUBNODES) + 1);
            parent.setAt(index, org.mwg.Type.ENODE, node);
            return node;
          }
          private static subInsert(parent: org.mwg.struct.ENode, key: Float64Array, value: number, strategyType: number, min: Float64Array, max: Float64Array, center: Float64Array, resolution: Float64Array, buffersize: number, root: org.mwg.struct.ENode, bufferupdate: boolean): boolean {
            let index: number = org.mwg.structure.trees.NDTree.getRelationId(center, key);
            let child: org.mwg.struct.ENode = <org.mwg.struct.ENode>parent.getAt(index);
            if (child == null) {
              child = org.mwg.structure.trees.NDTree.createNewNode(parent, root, index, min, max, center, key, buffersize);
            }
            let childmin: Float64Array = <Float64Array>child.getAt(NDTree.E_MIN);
            let childmax: Float64Array = <Float64Array>child.getAt(NDTree.E_MAX);
            let childcenter: Float64Array = org.mwg.structure.trees.NDTree.getCenterMinMax(childmin, childmax);
            let res: boolean = org.mwg.structure.trees.NDTree.internalInsert(child, key, value, strategyType, childmin, childmax, childcenter, resolution, buffersize, root);
            res = res && !bufferupdate;
            if (res) {
              switch (strategyType) {
                case org.mwg.structure.trees.IndexStrategy.PROFILE: {
                  parent.setAt(NDTree.E_TOTAL, org.mwg.Type.LONG, <number>parent.getAt(NDTree.E_TOTAL) + <number>value);
break;
                }
                case org.mwg.structure.trees.IndexStrategy.INDEX: {
                  parent.setAt(NDTree.E_TOTAL, org.mwg.Type.LONG, <number>parent.getAt(NDTree.E_TOTAL) + 1);
break;
                }
                default: 
 {
                  throw new Error("Index strategy is wrong!");
                }              }
            }
            return res;
          }
          private static internalInsert(node: org.mwg.struct.ENode, key: Float64Array, value: number, strategyType: number, min: Float64Array, max: Float64Array, center: Float64Array, resolution: Float64Array, buffersize: number, root: org.mwg.struct.ENode): boolean {
            if (<number>node.getAt(NDTree.E_SUBNODES) != 0) {
              return org.mwg.structure.trees.NDTree.subInsert(node, key, value, strategyType, min, max, center, resolution, buffersize, root, false);
            } else if (org.mwg.structure.trees.NDTree.checkCreateLevels(min, max, resolution)) {
              let buffer: org.mwg.struct.DMatrix = null;
              if (buffersize > 0) {
                buffer = <org.mwg.struct.DMatrix>node.getOrCreateAt(NDTree.E_BUFFER_KEYS, org.mwg.Type.DMATRIX);
              }
              if (buffer != null) {
                for (let i: number = 0; i < buffer.columns(); i++) {
                  if (org.mwg.structure.trees.NDTree.compare(key, buffer.column(i), resolution)) {
                    switch (strategyType) {
                      case org.mwg.structure.trees.IndexStrategy.PROFILE: {
                        let bufferkeys: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>node.getAt(NDTree.E_PROFILE);
                        for (let j: number = 0; j < key.length; j++) {
                          bufferkeys.set(j, i, bufferkeys.get(j, i) + key[j] * value);
                        }
                        let bufferValue: org.mwg.struct.LMatrix = <org.mwg.struct.LMatrix>node.getAt(NDTree.E_BUFFER_VALUES);
                        bufferValue.set(0, i, bufferValue.get(0, i) + value);
                        node.setAt(NDTree.E_TOTAL, org.mwg.Type.LONG, <number>node.getAt(NDTree.E_TOTAL) + value);
                        return true;
                      }
                      case org.mwg.structure.trees.IndexStrategy.INDEX: {
                        let bufferValue: org.mwg.struct.LMatrix = <org.mwg.struct.LMatrix>node.getAt(NDTree.E_BUFFER_VALUES);
                        bufferValue.set(0, i, value);
                        return false;
                      }
                      default: 
 {
                        throw new Error("Index strategy is wrong!");
                      }                    }
                  }
                }
                if (buffer.columns() < buffersize) {
                  buffer.appendColumn(key);
                  switch (strategyType) {
                    case org.mwg.structure.trees.IndexStrategy.PROFILE: {
                      let bufferkeys: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>node.getOrCreateAt(NDTree.E_PROFILE, org.mwg.Type.DMATRIX);
                      bufferkeys.appendColumn(key);
                      let bufferValue: org.mwg.struct.LMatrix = <org.mwg.struct.LMatrix>node.getOrCreateAt(NDTree.E_BUFFER_VALUES, org.mwg.Type.LMATRIX);
                      bufferValue.appendColumn(new Float64Array([value]));
                      node.setAt(NDTree.E_TOTAL, org.mwg.Type.LONG, <number>node.getAt(NDTree.E_TOTAL) + value);
                      return true;
                    }
                    case org.mwg.structure.trees.IndexStrategy.INDEX: {
                      let bufferValue: org.mwg.struct.LMatrix = <org.mwg.struct.LMatrix>node.getOrCreateAt(NDTree.E_BUFFER_VALUES, org.mwg.Type.LMATRIX);
                      bufferValue.appendColumn(new Float64Array([<number>value]));
                      node.setAt(NDTree.E_TOTAL, org.mwg.Type.LONG, <number>node.getAt(NDTree.E_TOTAL) + 1);
                      return true;
                    }
                    default: 
 {
                      throw new Error("Index strategy is wrong!");
                    }                  }
                } else {
                  if (strategyType == org.mwg.structure.trees.IndexStrategy.PROFILE) {
                    let bufferkeys: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>node.getAt(NDTree.E_PROFILE);
                    let bufferValue: org.mwg.struct.LMatrix = <org.mwg.struct.LMatrix>node.getAt(NDTree.E_BUFFER_VALUES);
                    for (let i: number = 0; i < buffer.columns(); i++) {
                      let t: number = bufferValue.get(0, i);
                      for (let j: number = 0; j < buffer.rows(); j++) {
                        buffer.set(j, i, bufferkeys.get(j, i) / t);
                      }
                    }
                    node.setAt(NDTree.E_PROFILE, org.mwg.Type.DMATRIX, null);
                  }
                  let bufferValue: org.mwg.struct.LMatrix = <org.mwg.struct.LMatrix>node.getAt(NDTree.E_BUFFER_VALUES);
                  for (let i: number = 0; i < buffer.columns(); i++) {
                    org.mwg.structure.trees.NDTree.subInsert(node, buffer.column(i), bufferValue.get(0, i), strategyType, min, max, center, resolution, buffersize, root, true);
                  }
                  node.setAt(NDTree.E_BUFFER_VALUES, org.mwg.Type.LMATRIX, null);
                  node.setAt(NDTree.E_BUFFER_KEYS, org.mwg.Type.DMATRIX, null);
                  return org.mwg.structure.trees.NDTree.subInsert(node, key, value, strategyType, min, max, center, resolution, buffersize, root, false);
                }
              } else {
                return org.mwg.structure.trees.NDTree.subInsert(node, key, value, strategyType, min, max, center, resolution, buffersize, root, false);
              }
            } else {
              switch (strategyType) {
                case org.mwg.structure.trees.IndexStrategy.PROFILE: {
                  let profile: Float64Array = <Float64Array>node.getAt(NDTree.E_PROFILE);
                  if (profile == null) {
                    profile = new Float64Array(key.length);
                    java.lang.System.arraycopy(key, 0, profile, 0, key.length);
                  } else {
                    for (let i: number = 0; i < key.length; i++) {
                      profile[i] += key[i] * value;
                    }
                  }
                  node.setAt(NDTree.E_PROFILE, org.mwg.Type.DOUBLE_ARRAY, profile);
                  node.setAt(NDTree.E_TOTAL, org.mwg.Type.LONG, <number>node.getAt(NDTree.E_TOTAL) + value);
                  return true;
                }
                case org.mwg.structure.trees.IndexStrategy.INDEX: {
                  if (<number>node.getAt(NDTree.E_TOTAL) == 0) {
                    node.setAt(NDTree.E_PROFILE, org.mwg.Type.DOUBLE_ARRAY, key);
                    node.setAt(NDTree.E_VALUE, org.mwg.Type.LONG, value);
                    node.setAt(NDTree.E_TOTAL, org.mwg.Type.LONG, 1);
                    return true;
                  } else {
                    node.setAt(NDTree.E_PROFILE, org.mwg.Type.DOUBLE_ARRAY, key);
                    node.setAt(NDTree.E_VALUE, org.mwg.Type.LONG, value);
                    return false;
                  }
                }
                default: 
 {
                  throw new Error("Index strategy is wrong!");
                }              }
            }

          }
          private static compare(key1: Float64Array, key2: Float64Array, resolution: Float64Array): boolean {
            if (resolution != null) {
              for (let i: number = 0; i < key1.length; i++) {
                if (Math.abs(key1[i] - key2[i]) > resolution[i]) {
                  return false;
                }
              }
            } else {
              for (let i: number = 0; i < key1.length; i++) {
                if (Math.abs(key1[i] - key2[i]) > 0) {
                  return false;
                }
              }
            }
            return true;
          }
          public setDistance(distanceType: number): void {
            super.set(NDTree.DISTANCE, org.mwg.Type.INT, distanceType);
          }
          public insert(keys: Float64Array, value: number): void {
            let state: org.mwg.plugin.NodeState = this.unphasedState();
            let min: Float64Array = <Float64Array>state.getFromKey(NDTree.BOUND_MIN);
            let max: Float64Array = <Float64Array>state.getFromKey(NDTree.BOUND_MAX);
            org.mwg.structure.trees.NDTree.check(keys, min, max);
            let resolution: Float64Array = <Float64Array>state.getFromKey(NDTree.RESOLUTION);
            let graph: org.mwg.struct.EGraph = <org.mwg.struct.EGraph>state.getOrCreateFromKey(NDTree.EGRAPH, org.mwg.Type.EGRAPH);
            let buffersize: number = state.getFromKeyWithDefault(NDTree.BUFFER_SIZE, NDTree.BUFFER_SIZE_DEF);
{
              let root: org.mwg.struct.ENode = graph.root();
              if (root == null) {
                root = graph.newNode();
                state.setFromKey(NDTree.STRATEGY, org.mwg.Type.INT, org.mwg.structure.trees.IndexStrategy.INDEX);
                graph.setRoot(root);
                root.setAt(NDTree.E_TOTAL, org.mwg.Type.LONG, 0);
                root.setAt(NDTree.E_TOTAL_SUBNODES, org.mwg.Type.LONG, 0);
                root.setAt(NDTree.E_SUBNODES, org.mwg.Type.LONG, 0);
                root.setAt(NDTree.E_MIN, org.mwg.Type.DOUBLE_ARRAY, min);
                root.setAt(NDTree.E_MAX, org.mwg.Type.DOUBLE_ARRAY, max);
              }
              org.mwg.structure.trees.NDTree.internalInsert(root, keys, value, org.mwg.structure.trees.IndexStrategy.INDEX, min, max, org.mwg.structure.trees.NDTree.getCenterMinMax(min, max), resolution, buffersize, root);
            }          }
          public profile(keys: Float64Array, occurrence: number): void {
            let state: org.mwg.plugin.NodeState = this.unphasedState();
            let min: Float64Array = <Float64Array>state.getFromKey(NDTree.BOUND_MIN);
            let max: Float64Array = <Float64Array>state.getFromKey(NDTree.BOUND_MAX);
            org.mwg.structure.trees.NDTree.check(keys, min, max);
            let resolution: Float64Array = <Float64Array>state.getFromKey(NDTree.RESOLUTION);
            let graph: org.mwg.struct.EGraph = <org.mwg.struct.EGraph>state.getOrCreateFromKey(NDTree.EGRAPH, org.mwg.Type.EGRAPH);
            let buffersize: number = state.getFromKeyWithDefault(NDTree.BUFFER_SIZE, NDTree.BUFFER_SIZE_DEF);
{
              let root: org.mwg.struct.ENode = graph.root();
              if (root == null) {
                root = graph.newNode();
                state.setFromKey(NDTree.STRATEGY, org.mwg.Type.INT, org.mwg.structure.trees.IndexStrategy.PROFILE);
                graph.setRoot(root);
                root.setAt(NDTree.E_TOTAL, org.mwg.Type.LONG, 0);
                root.setAt(NDTree.E_TOTAL_SUBNODES, org.mwg.Type.LONG, 0);
                root.setAt(NDTree.E_SUBNODES, org.mwg.Type.LONG, 0);
                root.setAt(NDTree.E_MIN, org.mwg.Type.DOUBLE_ARRAY, min);
                root.setAt(NDTree.E_MAX, org.mwg.Type.DOUBLE_ARRAY, max);
              }
              org.mwg.structure.trees.NDTree.internalInsert(root, keys, occurrence, org.mwg.structure.trees.IndexStrategy.PROFILE, min, max, org.mwg.structure.trees.NDTree.getCenterMinMax(min, max), resolution, buffersize, root);
            }          }
          public nearestN(keys: Float64Array, nbElem: number): org.mwg.structure.TreeResult {
            return this.nearestNWithinRadius(keys, -1, -1);
          }
          public nearestWithinRadius(keys: Float64Array, radius: number): org.mwg.structure.TreeResult {
            return this.nearestNWithinRadius(keys, -1, radius);
          }
          public nearestNWithinRadius(keys: Float64Array, nbElem: number, radius: number): org.mwg.structure.TreeResult {
            let state: org.mwg.plugin.NodeState = this.unphasedState();
            let min: Float64Array = <Float64Array>state.getFromKey(NDTree.BOUND_MIN);
            let max: Float64Array = <Float64Array>state.getFromKey(NDTree.BOUND_MAX);
            org.mwg.structure.trees.NDTree.check(keys, min, max);
            let graph: org.mwg.struct.EGraph = <org.mwg.struct.EGraph>state.getOrCreateFromKey(NDTree.EGRAPH, org.mwg.Type.EGRAPH);
            let distance: org.mwg.structure.distance.Distance = org.mwg.structure.distance.Distances.getDistance(state.getFromKeyWithDefault(NDTree.DISTANCE, NDTree.DISTANCE_DEF));
            let strategyType: number = <number>state.getFromKey(NDTree.STRATEGY);
            let root: org.mwg.struct.ENode = graph.root();
            if (root == null) {
              return null;
            }
            let calcZone: org.mwg.struct.EGraph = this.graph().space().newVolatileGraph();
            let nnl: org.mwg.structure.util.VolatileResult = new org.mwg.structure.util.VolatileResult(calcZone.newNode(), nbElem);
            org.mwg.structure.trees.NDTree.reccursiveTraverse(root, calcZone, nnl, strategyType, distance, keys, null, null, null, radius);
            nnl.sort(true);
            return nnl;
          }
          public query(min: Float64Array, max: Float64Array): org.mwg.structure.TreeResult {
            let state: org.mwg.plugin.NodeState = this.unphasedState();
            let graph: org.mwg.struct.EGraph = <org.mwg.struct.EGraph>state.getOrCreateFromKey(NDTree.EGRAPH, org.mwg.Type.EGRAPH);
            let distance: org.mwg.structure.distance.Distance = org.mwg.structure.distance.Distances.getDistance(state.getFromKeyWithDefault(NDTree.DISTANCE, NDTree.DISTANCE_DEF));
            let strategyType: number = <number>state.getFromKey(NDTree.STRATEGY);
            let center: Float64Array = new Float64Array(max.length);
            for (let i: number = 0; i < center.length; i++) {
              center[i] = (min[i] + max[i]) / 2;
            }
            let root: org.mwg.struct.ENode = graph.root();
            if (root == null) {
              return null;
            }
            let calcZone: org.mwg.struct.EGraph = this.graph().space().newVolatileGraph();
            let nnl: org.mwg.structure.util.VolatileResult = new org.mwg.structure.util.VolatileResult(calcZone.newNode(), -1);
            org.mwg.structure.trees.NDTree.reccursiveTraverse(root, calcZone, nnl, strategyType, distance, null, min, max, center, -1);
            nnl.sort(true);
            return nnl;
          }
          public size(): number {
            let state: org.mwg.plugin.NodeState = this.unphasedState();
            let graph: org.mwg.struct.EGraph = <org.mwg.struct.EGraph>state.getOrCreateFromKey(NDTree.EGRAPH, org.mwg.Type.EGRAPH);
            let root: org.mwg.struct.ENode = graph.root();
            if (root == null) {
              return 0;
            } else {
              return <number>root.getAt(NDTree.E_TOTAL);
            }
          }
          public numberOfNodes(): number {
            let state: org.mwg.plugin.NodeState = this.unphasedState();
            let graph: org.mwg.struct.EGraph = <org.mwg.struct.EGraph>state.getOrCreateFromKey(NDTree.EGRAPH, org.mwg.Type.EGRAPH);
            let root: org.mwg.struct.ENode = graph.root();
            if (root == null) {
              return 0;
            } else {
              return <number>root.getAt(NDTree.E_TOTAL_SUBNODES);
            }
          }
          private static binaryFromLong(value: number, dim: number): boolean[] {
            let tempvalue: number = value - NDTree.E_OFFSET_REL;
            let shiftvalue: number = tempvalue >> 1;
            let res: boolean[] = [];
            for (let i: number = 0; i < dim; i++) {
              res[dim - i - 1] = ((tempvalue - (shiftvalue << 1)) == 1);
              tempvalue = shiftvalue;
              shiftvalue = tempvalue >> 1;
            }
            return res;
          }
          private static reccursiveTraverse(node: org.mwg.struct.ENode, calcZone: org.mwg.struct.EGraph, nnl: org.mwg.structure.util.VolatileResult, strategyType: number, distance: org.mwg.structure.distance.Distance, target: Float64Array, targetmin: Float64Array, targetmax: Float64Array, targetcenter: Float64Array, radius: number): void {
            if (<number>node.getAt(NDTree.E_SUBNODES) == 0) {
              let buffer: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>node.getAt(NDTree.E_BUFFER_KEYS);
              let bufferValue: org.mwg.struct.LMatrix = <org.mwg.struct.LMatrix>node.getAt(NDTree.E_BUFFER_VALUES);
              if (buffer != null) {
                switch (strategyType) {
                  case org.mwg.structure.trees.IndexStrategy.PROFILE: {
                    let tempK: Float64Array = new Float64Array(target.length);
                    let bufferkeys: org.mwg.struct.DMatrix = <org.mwg.struct.DMatrix>node.getAt(NDTree.E_PROFILE);
                    for (let i: number = 0; i < buffer.columns(); i++) {
                      let t: number = bufferValue.get(0, i);
                      for (let j: number = 0; j < buffer.rows(); j++) {
                        tempK[j] = bufferkeys.get(j, i) / t;
                      }
                      org.mwg.structure.util.TreeHelper.filterAndInsert(tempK, t, target, targetmin, targetmax, targetcenter, distance, radius, nnl);
                    }
                    return;
                  }
                  case org.mwg.structure.trees.IndexStrategy.INDEX: {
                    for (let i: number = 0; i < buffer.columns(); i++) {
                      org.mwg.structure.util.TreeHelper.filterAndInsert(buffer.column(i), bufferValue.get(0, i), target, targetmin, targetmax, targetcenter, distance, radius, nnl);
                    }
                    return;
                  }
                  default: 
 {
                    throw new Error("Index strategy is wrong!");
                  }                }
              } else {
                switch (strategyType) {
                  case org.mwg.structure.trees.IndexStrategy.PROFILE: {
                    let keyo: Float64Array = <Float64Array>node.getAt(NDTree.E_PROFILE);
                    let key: Float64Array = new Float64Array(keyo.length);
                    let value: number = <number>node.getAt(NDTree.E_TOTAL);
                    for (let i: number = 0; i < keyo.length; i++) {
                      key[i] = keyo[i] / value;
                    }
                    org.mwg.structure.util.TreeHelper.filterAndInsert(key, value, target, targetmin, targetmax, targetcenter, distance, radius, nnl);
                    return;
                  }
                  case org.mwg.structure.trees.IndexStrategy.INDEX: {
                    let key: Float64Array = <Float64Array>node.getAt(NDTree.E_PROFILE);
                    let value: number = <number>node.getAt(NDTree.E_VALUE);
                    org.mwg.structure.util.TreeHelper.filterAndInsert(key, value, target, targetmin, targetmax, targetcenter, distance, radius, nnl);
                    return;
                  }
                  default: 
 {
                    throw new Error("Index strategy is wrong!");
                  }                }
              }
            } else {
              let boundMax: Float64Array = <Float64Array>node.getAt(NDTree.E_MAX);
              let boundMin: Float64Array = <Float64Array>node.getAt(NDTree.E_MIN);
              let worst: number = nnl.getWorstDistance();
              if (targetmin == null || targetmax == null) {
                if (!nnl.isCapacityReached() || org.mwg.structure.util.TreeHelper.getclosestDistance(target, boundMin, boundMax, distance) <= worst) {
                  let tempList: org.mwg.struct.ENode = calcZone.newNode();
                  let childPriority: org.mwg.structure.util.VolatileResult = new org.mwg.structure.util.VolatileResult(tempList, -1);
                  let dim: number = boundMax.length;
                  let childMin: Float64Array = new Float64Array(dim);
                  let childMax: Float64Array = new Float64Array(dim);
                  node.each((attributeKey : number, elemType : number, elem : any) => {
{
                      if (attributeKey >= org.mwg.structure.trees.NDTree.E_OFFSET_REL) {
                        let binaries: boolean[] = org.mwg.structure.trees.NDTree.binaryFromLong(attributeKey, dim);
                        for (let i: number = 0; i < dim; i++) {
                          if (!binaries[i]) {
                            childMin[i] = boundMin[i];
                            childMax[i] = (boundMax[i] + boundMin[i]) / 2;
                          } else {
                            childMin[i] = (boundMax[i] + boundMin[i]) / 2;
                            childMax[i] = boundMax[i];
                          }
                        }
                        childPriority.insert(childMin, attributeKey, org.mwg.structure.util.TreeHelper.getclosestDistance(target, childMin, childMax, distance));
                      }
                    }                  });
                  childPriority.sort(true);
                  for (let i: number = 0; i < childPriority.size(); i++) {
                    let child: org.mwg.struct.ENode = <org.mwg.struct.ENode>node.getAt(<number>childPriority.value(i));
                    org.mwg.structure.trees.NDTree.reccursiveTraverse(child, calcZone, nnl, strategyType, distance, target, targetmin, targetmax, targetcenter, radius);
                  }
                  childPriority.free();
                }
              } else {
                node.each((attributeKey : number, elemType : number, elem : any) => {
{
                    if (attributeKey >= org.mwg.structure.trees.NDTree.E_OFFSET_REL) {
                      let child: org.mwg.struct.ENode = <org.mwg.struct.ENode>node.getAt(attributeKey);
                      if (org.mwg.structure.util.TreeHelper.checkBoundsIntersection(targetmin, targetmax, <Float64Array>child.getAt(org.mwg.structure.trees.NDTree.E_MIN), <Float64Array>child.getAt(org.mwg.structure.trees.NDTree.E_MAX))) {
                        org.mwg.structure.trees.NDTree.reccursiveTraverse(child, calcZone, nnl, strategyType, distance, target, targetmin, targetmax, targetcenter, radius);
                      }
                    }
                  }                });
              }
            }
          }
        }
      }
      export module util {
        export class HRect {
          public min: Float64Array;
          public max: Float64Array;
          constructor(vmin: Float64Array, vmax: Float64Array) {
            this.min = new Float64Array(vmin.length);
            this.max = new Float64Array(vmax.length);
            java.lang.System.arraycopy(vmin, 0, this.min, 0, vmin.length);
            java.lang.System.arraycopy(vmax, 0, this.max, 0, vmax.length);
          }
          public clone(): any {
            return new org.mwg.structure.util.HRect(this.min, this.max);
          }
          public closest(t: Float64Array): Float64Array {
            let p: Float64Array = new Float64Array(t.length);
            for (let i: number = 0; i < t.length; ++i) {
              if (t[i] <= this.min[i]) {
                p[i] = this.min[i];
              } else if (t[i] >= this.max[i]) {
                p[i] = this.max[i];
              } else {
                p[i] = t[i];
              }

            }
            return p;
          }
          public static infiniteHRect(d: number): org.mwg.structure.util.HRect {
            let vmin: Float64Array = new Float64Array(d);
            let vmax: Float64Array = new Float64Array(d);
            for (let i: number = 0; i < d; ++i) {
              vmin[i] = java.lang.Double.NEGATIVE_INFINITY;
              vmax[i] = java.lang.Double.POSITIVE_INFINITY;
            }
            return new org.mwg.structure.util.HRect(vmin, vmax);
          }
          public toString(): string {
            return this.min + "\n" + this.max + "\n";
          }
        }
        export class TreeHelper {
          public static checkBoundsIntersection(targetmin: Float64Array, targetmax: Float64Array, boundMin: Float64Array, boundMax: Float64Array): boolean {
            for (let i: number = 0; i < boundMax.length; i++) {
              if (targetmin[i] > boundMax[i] || targetmax[i] < boundMin[i]) {
                return false;
              }
            }
            return true;
          }
          public static checkKeyInsideBounds(key: Float64Array, boundMin: Float64Array, boundMax: Float64Array): boolean {
            for (let i: number = 0; i < boundMax.length; i++) {
              if (key[i] > boundMax[i] || key[i] < boundMin[i]) {
                return false;
              }
            }
            return true;
          }
          public static getclosestDistance(target: Float64Array, boundMin: Float64Array, boundMax: Float64Array, distance: org.mwg.structure.distance.Distance): number {
            let closest: Float64Array = new Float64Array(target.length);
            for (let i: number = 0; i < target.length; i++) {
              if (target[i] >= boundMax[i]) {
                closest[i] = boundMax[i];
              } else if (target[i] <= boundMin[i]) {
                closest[i] = boundMin[i];
              } else {
                closest[i] = target[i];
              }

            }
            return distance.measure(closest, target);
          }
          public static filterAndInsert(key: Float64Array, value: number, target: Float64Array, targetmin: Float64Array, targetmax: Float64Array, targetcenter: Float64Array, distance: org.mwg.structure.distance.Distance, radius: number, nnl: org.mwg.structure.util.VolatileResult): void {
            if (targetmin != null) {
              if (org.mwg.structure.util.TreeHelper.checkKeyInsideBounds(key, targetmin, targetmax)) {
                nnl.insert(key, value, distance.measure(key, targetcenter));
              }
            } else {
              let dist: number = distance.measure(key, target);
              if (radius > 0) {
                if (dist < radius) {
                  nnl.insert(key, value, dist);
                }
              } else {
                nnl.insert(key, value, dist);
              }
            }
          }
        }
        export class VolatileResult implements org.mwg.structure.TreeResult {
          private static maxPriority: number = java.lang.Double.MAX_VALUE;
          private static _KEYS: number = 1;
          private static _VALUES: number = 2;
          private static _DISTANCES: number = 3;
          private node: org.mwg.struct.ENode;
          private capacity: number;
          private count: number;
          private worst: number;
          private _keys: org.mwg.struct.DMatrix;
          private _values: org.mwg.struct.LMatrix;
          private _distances: org.mwg.struct.DMatrix;
          constructor(node: org.mwg.struct.ENode, capacity: number) {
            this.node = node;
            this.count = 0;
            this._keys = <org.mwg.struct.DMatrix>node.getOrCreateAt(VolatileResult._KEYS, org.mwg.Type.DMATRIX);
            this._values = <org.mwg.struct.LMatrix>node.getOrCreateAt(VolatileResult._VALUES, org.mwg.Type.LMATRIX);
            this._distances = <org.mwg.struct.DMatrix>node.getOrCreateAt(VolatileResult._DISTANCES, org.mwg.Type.DMATRIX);
            this.capacity = capacity;
          }
          public size(): number {
            return this.count;
          }
          public insert(key: Float64Array, value: number, distance: number): boolean {
            if (this.capacity > 0 && this.count == this.capacity) {
              this.add(key, value, distance, true);
              return true;
            }
            this.add(key, value, distance, false);
            return true;
          }
          private add(key: Float64Array, value: number, distance: number, remove: boolean): void {
            if (this.count == 0) {
              this._keys.appendColumn(new Float64Array(key.length));
              this._values.appendColumn(new Float64Array(1));
              this._distances.appendColumn(new Float64Array([VolatileResult.maxPriority]));
            }
            if (remove) {
              if (distance > this.getWorstDistance()) {
                return;
              }
              this.remove();
              this.count++;
              for (let i: number = 0; i < this._keys.rows(); i++) {
                this._keys.set(i, this.count, key[i]);
              }
              this._values.set(0, this.count, value);
              this._distances.set(0, this.count, distance);
            } else {
              this.count++;
              this._keys.appendColumn(key);
              this._values.appendColumn(new Float64Array([value]));
              this._distances.appendColumn(new Float64Array([distance]));
            }
            this.bubbleUp(this.count);
            this.worst = this._distances.get(0, 1);
          }
          private remove(): void {
            if (this.count == 0)
              return;

            for (let i: number = 0; i < this._keys.rows(); i++) {
              this._keys.set(i, 1, this._keys.get(i, this.count));
              this._keys.set(i, this.count, 0);
            }
            this._values.set(0, 1, this._values.get(0, this.count));
            this._distances.set(0, 1, this._distances.get(0, this.count));
            this._values.set(0, this.count, 0);
            this._distances.set(0, this.count, 0);
            this.count--;
            this.bubbleDown(1);
          }
          private bubbleUp(pos: number): void {
            let okey: Float64Array = this._keys.column(pos);
            let element: number = this._values.column(pos)[0];
            let priority: number = this._distances.column(pos)[0];
            let halfpos: number = <number>Math.floor(pos / 2);
            while (this._distances.column(halfpos)[0] < priority) {
              this._distances.set(0, pos, this._distances.get(0, halfpos));
              this._values.set(0, pos, this._values.get(0, halfpos));
              for (let i: number = 0; i < this._keys.rows(); i++) {
                this._keys.set(i, pos, this._keys.get(i, halfpos));
              }
              pos = <number>Math.floor(pos / 2);
              halfpos = <number>Math.floor(pos / 2);
            }
            this._distances.set(0, pos, priority);
            this._values.set(0, pos, element);
            for (let i: number = 0; i < this._keys.rows(); i++) {
              this._keys.set(i, pos, okey[i]);
            }
          }
          private bubbleDown(pos: number): void {
            let okey: Float64Array = this._keys.column(pos);
            let element: number = this._values.column(pos)[0];
            let priority: number = this._distances.column(pos)[0];
            let child: number;
            for (; pos * 2 <= this.count; pos = child) {
              child = pos * 2;
              if (child != this.count)
                if (this._distances.get(0, child) < this._distances.get(0, child + 1))
                  child++;


              if (priority < this._distances.get(0, child)) {
                this._distances.set(0, pos, this._distances.get(0, child));
                this._values.set(0, pos, this._values.get(0, child));
                for (let i: number = 0; i < this._keys.rows(); i++) {
                  this._keys.set(i, pos, this._keys.get(i, child));
                }
              } else {
break;
              }
            }
            this._distances.set(0, pos, priority);
            this._values.set(0, pos, element);
            for (let i: number = 0; i < this._keys.rows(); i++) {
              this._keys.set(i, pos, okey[i]);
            }
          }
          public keys(index: number): Float64Array {
            return this._keys.column(index + 1);
          }
          public value(index: number): number {
            return this._values.get(index + 1, 0);
          }
          public distance(index: number): number {
            return this._distances.get(index + 1, 0);
          }
          public getWorstDistance(): number {
            if (this.count == 0) {
              return java.lang.Double.POSITIVE_INFINITY;
            }
            return this.worst;
          }
          public free(): void {
            let g: org.mwg.struct.EGraph = this.node.graph();
            this.node.drop();
            if (g.size() == 0) {
              g.free();
            }
          }
          public isCapacityReached(): boolean {
            return this.capacity > 0 && (this.count == this.capacity);
          }
          private swap(i: number, j: number): void {
            let tempkey: Float64Array = this._keys.column(i);
            let tempvalue: number = this._values.get(0, i);
            let tempdist: number = this._distances.get(0, i);
            this._distances.set(0, i, this._distances.get(0, j));
            this._values.set(0, i, this._values.get(0, j));
            for (let k: number = 0; k < this._keys.rows(); k++) {
              this._keys.set(k, i, this._keys.get(k, j));
            }
            this._distances.set(0, j, tempdist);
            this._values.set(0, j, tempvalue);
            for (let k: number = 0; k < this._keys.rows(); k++) {
              this._keys.set(k, j, tempkey[k]);
            }
          }
          private partition(l: number, h: number, ascending: boolean): number {
            let x: number = this._distances.get(0, h);
            let i: number = (l - 1);
            for (let j: number = l; j <= h - 1; j++) {
              if (ascending) {
                if (this._distances.get(0, j) <= x) {
                  i++;
                  this.swap(i, j);
                }
              } else {
                if (this._distances.get(0, j) > x) {
                  i++;
                  this.swap(i, j);
                }
              }
            }
            this.swap(i + 1, h);
            return (i + 1);
          }
          private quickSort(l: number, h: number, ascending: boolean): void {
            let stack: Int32Array = new Int32Array(h - l + 1);
            let top: number = -1;
            stack[++top] = l;
            stack[++top] = h;
            while (top >= 0) {
              h = stack[top--];
              l = stack[top--];
              let p: number = this.partition(l, h, ascending);
              if (p - 1 > l) {
                stack[++top] = l;
                stack[++top] = p - 1;
              }
              if (p + 1 < h) {
                stack[++top] = p + 1;
                stack[++top] = h;
              }
            }
          }
          public sort(ascending: boolean): void {
            if (this.count > 1) {
              this.quickSort(1, this.count, ascending);
            }
          }
        }
      }
    }
  }
}
